% Example of NEGA method for linear time-invariant model
% - scalar example
% - supplementary code for submitted article "Covariance Estimation and Gaussianity Assessment 
%   for State and Measurement Noise"
% Authors: J. Dunik, O. Kost, O. Straka, E. Blasch
% Corresponding author: J. Dunik (dunikj@kky.zcu.cz)

% MODEL DESCRIPTION AND DATA SIMULATION
% - selection of Gaussian and non-Gaussian measurement noise (1 = Gaussian, 2 = Student's t)
measurementNoiseGaussian = 1;
% - dimensions
nx = 1; % state
nz = 1; % measurement
% - number of data
N = 1e4;
% - model matrices
F = 0.5;
H = 2;
% - variances of noises (specified just for simulation)
Q = 1; % state
R = 2; % measurement
% - data generator
x = zeros(1,N+1); % state
z = zeros(1,N);   % measurement
for k = 1:N
    % - noises realisations    
    % -- state noise (assumed to be Gaussian)    
    wk = sqrt(Q)*randn;
    % -- measurement noise
    if measurementNoiseGaussian == 1
        % -- (assumed to be Gaussian)
        vk = sqrt(R)*randn;
    else
        % -- (assumed to be Student's t)
        c = 2*R/(R-1);
        vk = trnd(c);
    end
    % - state and measurement generation
    x(k+1) = F*x(k) + wk;
    z(k) = H*x(k) + vk;
end

% NEGA METHOD
% - Noise variances Q and R estimation
% -- definition of design parameters
L = 2;
Lp = L + 1;
% -- number of unknown and estimated variances (Q and R are scalars)
nTheta = 2;
% -- computation of observability and auxilliary matrices in (40), (41)
O = [H; H*F];
Gamma = [0 0; H 0];
% -- computation of augmented measurement prediction error 
Z = [z(:,1:end-1); z(:,2:end)]; % augmented measurement (14)
Zhat = zeros(L*nz,N-1);   % augmented measurement prediction (42)
Ztilde = zeros(L*nz,N-1); % augmented measurement prediction error (18)
for k = 2:N-1
    Zhat(:,k) = O*F*pinv(O)*Z(:,k-1);
    Ztilde(:,k) = Z(:,k) - Zhat(:,k);
end
% -- computation of matrices for LS solution (44)-(45)
% --- A matrix in (43)
Aw = [eye(L*nz), eye(L*nz)] * [[O, Gamma]; [-O*F*pinv(O)*Gamma, zeros(L*nz,nx)]];
Av = [eye(L*nz), eye(L*nz)] * [[zeros(L*nz,nx), eye(L*nz)]; [-O*F*pinv(O), zeros(L*nz,nx)]];
A = [Aw, Av];
% % --- SIGMA (26) - just for illustration, not necessary for NEGA
% %     method itself
% SIGMA = blkdiag(kron(eye(Lp),Q),kron(eye(Lp),R));
% SIGMAs = SIGMA(:);
% --- design of duplication matrix PSI in (29) and (32) (note that theta in (12) is [Q; R])
PSI = [1 0; zeros(6,2); 1 0; zeros(6,2); 1 0; zeros(6,2); 0 1; zeros(6,2); 0 1; zeros(6,2); 0 1];
% --- matrix DELTA in (47)
DELTA = kron(A,A)*PSI;
% -- construction of sample based estimate of vector b in (48), (49)
Chat = zeros(L*nz,L*nz);
for k = 2:N-1
    Chat = Chat + Ztilde(:,k)*(Ztilde(:,k))';
end
Chat = 1/(N-2)*Chat;
bHat = Chat(:);
% -- noise variances estimation by (50)
thetaHat = pinv(DELTA)*bHat;
Qhat = thetaHat(1);
Rhat = thetaHat(2);
% -- disp
disp(['True noise variances are: Q = ', num2str(Q), ', R = ', num2str(R)])
disp(['Estimated noise variances are: Qhat = ', num2str(Qhat), ', Rhat = ', num2str(Rhat)])

% - Gaussianity assessment
% -- definition of significance level
alpha = 0.01;
% -- calculation of transformed augmented prediction error (stochastic
%    decoupling) in (56) is not necessary
% -- decision about Gaussianity of decoupled augmented measurement prediction error
validHypothesis_element1 = jbtest(Ztilde(1,2:N-1),alpha/2);
validHypothesis_element2 = jbtest(Ztilde(2,2:N-1),alpha/2);
validHypothesis = validHypothesis_element1 & validHypothesis_element2;
% disp
if validHypothesis == 0
    disp('Null hypothesis H0 is accepted, i.e., noises are Gaussian.')
else
    disp('Null hypothesis H0 is rejected, i.e., noises are not Gaussian.')
end