% Example of NEGA method for linear time-varying model
% - scalar example
% - supplementary code for submitted article "Covariance Estimation and Gaussianity Assessment 
%   for State and Measurement Noise"
% Authors: J. Dunik, O. Kost, O. Straka, E. Blasch
% Corresponding author: J. Dunik (dunikj@kky.zcu.cz)

% MODEL DESCRIPTION AND DATA SIMULATION
% - selection of Gaussian and non-Gaussian measurement noise (1 = Gaussian, 2 = Student's t)
measurementNoiseGaussian = 1;
% - dimensions
nx = 1; % state
nz = 1; % measurement
% - number of data
N = 1e4;
% - model matrices
F = zeros(nx,N); % state
H = zeros(nx,N); % measurement
for k = 1:N
    F(k) = 0.5 + 0.4*sin(2*pi*k/N);
    H(k) = 2 + sin(10*pi*k/N);
end 
% - variances of noises (specified just for simulation)
Q = 1; % state
R = 2; % measurement
% - data generator
x = zeros(1,N+1); % state
z = zeros(1,N);   % measurement
for k = 1:N
    % - noises realisations    
    % -- state noise (assumed to be Gaussian)    
    wk = sqrt(Q)*randn;
    % -- measurement noise
    if measurementNoiseGaussian == 1
        % -- (assumed to be Gaussian)
        vk = sqrt(R)*randn;
    else
        % -- (assumed to be Student's t)
        c = 2*R/(R-1);
        vk = trnd(c);
    end
    % - state and measurement generation
    x(k+1) = F(k)*x(k) + wk;
    z(k) = H(k)*x(k) + vk;
end

% NEGA METHOD
% - Noise variances Q and R estimation
% -- definition of design parameters
L = 2;
Lp = L + 1;
% -- number of unknown and estimated variances (Q and R are scalars)
nTheta = 2;
% -- computation of observability and auxilliary matrices in (12), (15)
O = zeros(L*nz,nx,N-1);
Gamma = zeros(L*nz,L*nx,N-1);
for k = 1:N-1
    O(:,:,k) = [H(k); H(k+1)*F(k)];
    Gamma(:,:,k) = [0 0; H(k+1) 0];
end
% -- computation of augmented measurement prediction error 
Z = [z(:,1:end-1); z(:,2:end)]; % augmented measurement (14)
Zhat = zeros(L*nz,N-1);   % augmented measurement prediction (16)
Ztilde = zeros(L*nz,N-1); % augmented measurement prediction error (18)
for k = 2:N-1    
    Zhat(:,k) = O(:,:,k)*F(k-1)*pinv(O(:,:,k-1))*Z(:,k-1);
    Ztilde(:,k) = Z(:,k) - Zhat(:,k);
end
% -- computation of matrices for LS solution (38)
% --- A(k) matrix in (20)
Aw = zeros(L*nz,Lp*nx,N-1); % (23)
Av = zeros(L*nz,Lp*nz,N-1); % (24)
Ak = zeros(L*nz,Lp*(nx+nz),N-1); % (22)
for k = 2:N-1
    Aw(:,:,k) = [eye(L*nz), eye(L*nz)] * [[O(:,:,k), Gamma(:,:,k)]; [-O(:,:,k)*F(k-1)*pinv(O(:,:,k-1))*Gamma(:,:,k-1), zeros(L*nz,nx)]];
    Av(:,:,k) = [eye(L*nz), eye(L*nz)] * [[zeros(L*nz,nx), eye(L*nz)]; [-O(:,:,k)*F(k-1)*pinv(O(:,:,k-1)), zeros(L*nz,nx)]];
    Ak(:,:,k) = [Aw(:,:,k), Av(:,:,k)];
end
% % --- SIGMA (26) - just for illustration, not necessary for NEGA
% %     method itself
% SIGMA = blkdiag(kron(eye(Lp),Q),kron(eye(Lp),R));
% SIGMAs = SIGMA(:);
% --- design of duplication matrix PSI in (29) and (32) (note that theta in (12) is [Q; R])
PSI = [1 0; zeros(6,2); 1 0; zeros(6,2); 1 0; zeros(6,2); 0 1; zeros(6,2); 0 1; zeros(6,2); 0 1];
% --- matrix LAMBDA(k) (29)
LAMBDAk = zeros((L*nz)^2,nTheta,N-1);
for k = 2:N-1
    LAMBDAk(:,:,k) = kron(Ak(:,:,k),Ak(:,:,k))*PSI;
end
% --- matrix LAMBDA in (34)
LAMBDA = [];
for k = 2:N-1
    LAMBDA = [LAMBDA; LAMBDAk(:,:,k)]; %#ok<AGROW>
end
% -- construction of sample based estimate of vector b in (36), (37)
bHatk = zeros((L*nz)^2,N-1);
bHat = [];
for k = 2:N-1
    bHatk(:,k) = kron(Ztilde(:,k),Ztilde(:,k));
    bHat = [bHat; bHatk(:,k)]; %#ok<AGROW>
end
% -- noise variances estimation by (38)
thetaHat = pinv(LAMBDA)*bHat;
Qhat = thetaHat(1);
Rhat = thetaHat(2);
% -- disp
disp(['True noise variances are: Q = ', num2str(Q), ', R = ', num2str(R)])
disp(['Estimated noise variances are: Qhat = ', num2str(Qhat), ', Rhat = ', num2str(Rhat)])

% - Gaussianity assessment
% -- definition of significance level
alpha = 0.01;
% -- calculation of transformed augmented prediction error (stochastic
%    decoupling) in (56)
% --- estimate of augmented prediction error covariance
SIGMAhat = blkdiag(kron(eye(Lp),Qhat),kron(eye(Lp),Rhat));
CkHat = zeros(L*nz,L*nz,N-1);
SkHat = zeros(L*nz,L*nz,N-1);
for k = 2:N-1
    CkHat(:,:,k) = Ak(:,:,k)*SIGMAhat*Ak(:,:,k)';
    SkHat(:,:,k) = chol(CkHat(:,:,k));
end
% --- decoupled augmented measurement prediction error
ksi = zeros(L*nz,N-1);
for k = 2:N-1
    ksi(:,k) = inv(SkHat(:,:,k))*Ztilde(:,k); %#ok<MINV>
end
% -- decision about Gaussianity of decoupled augmented measurement prediction error
validHypothesis_element1 = jbtest(ksi(1,2:N-1),alpha/2);
validHypothesis_element2 = jbtest(ksi(2,2:N-1),alpha/2);
validHypothesis = validHypothesis_element1 & validHypothesis_element2;
% disp
if validHypothesis == 0
    disp('Null hypothesis H0 is accepted, i.e., noises are Gaussian.')
else
    disp('Null hypothesis H0 is rejected, i.e., noises are not Gaussian.')
end